package de.fzj.unicore.uas.admin;

import java.security.cert.X509Certificate;

import org.apache.log4j.Level;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import de.fzj.unicore.uas.UAS;
import de.fzj.unicore.uas.impl.registry.RegistryHandler;
import de.fzj.unicore.uas.impl.tss.TargetSystemHomeImpl;
import de.fzj.unicore.uas.util.LogUtil;
import de.fzj.unicore.uas.xnjs.XNJSFacade;
import de.fzj.unicore.wsrflite.Kernel;
import de.fzj.unicore.wsrflite.security.ISecurityProperties;
import de.fzj.unicore.wsrflite.security.SecurityManager;
import de.fzj.unicore.wsrflite.xmlbeans.client.ExternalRegistryClient;
import de.fzj.unicore.xnjs.legacy.LegacyTSI;
import de.fzj.unicore.xnjs.legacy.TSIConnectionFactory;
import de.fzj.unicore.xnjs.simple.LocalTS;
import de.fzj.unicore.xnjs.tsi.TSIFactory;

/**
 * general information and admin operations available via JMX
 */
public class UASAdmin implements UASAdminMBean {

	private static final Logger logger=LogUtil.getLogger(LogUtil.UNICORE,UASAdmin.class);

	public String getSiteName(){
		return UAS.getProperty("uas.targetsystem.sitename");
	}

	public String[] getExternalRegistryURLs(){
		String[] url=RegistryHandler.getExternalRegistryURLs();
		if(url==null || url.length==0)return new String[]{"<using internal registry>"};
		return url;
	}

	public String getIsSharedRegistry(){
		return Boolean.toString(RegistryHandler.isGlobalRegistry());
	}

	public String getProxyCertificateSupportEnabled(){
		return Kernel.getKernel().getProperty(ISecurityProperties.UAS_AUTHORISER_PROXY_SUPPORT, "false");
	}

	public String getAccessControlEnabled(){
		return Kernel.getKernel().getProperty(ISecurityProperties.UAS_CHECKACCESS);
	}

	public String getRequireSignatureEnabled(){
		return Kernel.getKernel().getProperty(ISecurityProperties.UAS_REQUIRE_SIGNATURES, "false");
	}

	public String getBaseURL(){
		return Kernel.getKernel().getProperty(Kernel.WSRF_BASEURL);  
	}

	public String getGatewayIdentity(){
		String id="<n/a>";
		if(Kernel.getKernel().getSecurityProperties().isSslEnabled()){
			X509Certificate c=SecurityManager.getGatewayCert();
			id=c.getSubjectX500Principal().getName();
		}
		return id;
	}

	public String setLogLevel(String log, String level){
		Level l=Level.toLevel(level);
		LogManager.getLogger(log).setLevel(l);
		logger.info("JMX ACTION: level for logger <"+log+"> changed to <"+l.toString()+">");
		return "Level set to "+l.toString();
	}


	public void setProperty(String key, String value){
		Kernel.getKernel().setProperty(key, value);
		logger.info("JMX ACTION: property <"+key+"> set to <"+value+">");
	}

	public String getGatewayConnectionStatus(){
		return Kernel.getKernel().getGatewayHandler().getConnectionStatusMessage();
	}

	public String getAttributeSourcesStatus(){
		try{
			return SecurityManager.getAuthoriserConnectionStatus();
		}catch(Exception ex){
			return "N/A (Exception occurred)";
		}
	}

	public String getTSIConnectionStatus(){
		try{
			if(Kernel.getKernel().getHome(UAS.TSS)==null){
				return "N/A";
			}
			String tsiClass=XNJSFacade.get(null).getConfiguration().getProperty(TSIFactory.TSI_TO_USE);
			if(LegacyTSI.class.getName().equals(tsiClass)){
				TSIConnectionFactory tsif=XNJSFacade.get(null).getConfiguration().getComponentInstanceOfType(TSIConnectionFactory.class);
				return tsif.getConnectionStatus();
			}
			else if(tsiClass==null || LocalTS.class.getName().equals(tsiClass)){
				return "OK [using embedded TSI]";
			}
		}catch(Exception e){}
		return "N/A";
	}

	public String getRegistryConnectionStatus(){
		if(!RegistryHandler.usesExternalRegistry()){
			return "N/A (no external registry used)";
		}
		else {
			try{
				String[]urls=RegistryHandler.getExternalRegistryURLs();
				ExternalRegistryClient erc=RegistryHandler.getExternalRegistryClient();
				if(erc.checkConnection()){
					String allurls="";
					for(String u: urls)allurls+=u+" ";
							return "OK [connected to "+allurls+"]";
				}
				else{
					if(urls.length>1){
						return "ONE OR MORE REGISTRIES ARE NOT AVAILABLE";
					}
					else return "CAN'T CONNECT TO REGISTRY";
				}
			}catch(Exception ex){
				return "Error! "+ex;
			}
		}
	}

	public String isJobSubmissionEnabled() {
		return String.valueOf(((TargetSystemHomeImpl)Kernel.getKernel().getHome(UAS.TSS)).isJobSubmissionEnabled());
	}

	public void toggleJobSubmissionEnabled() {
		TargetSystemHomeImpl th=(TargetSystemHomeImpl)Kernel.getKernel().getHome(UAS.TSS);
		boolean enabled=th.isJobSubmissionEnabled();
		th.setJobSubmissionEnabled(!enabled);
	}

}
