package de.fzj.unicore.uas.fts.http;

import java.io.IOException;
import java.math.BigInteger;
import java.util.Map;

import javax.xml.namespace.QName;

import org.apache.log4j.Logger;
import org.apache.xmlbeans.XmlObject;
import org.ggf.schemas.byteio.x2005.x10.randomAccess.SizeDocument;

import de.fzj.unicore.uas.UAS;
import de.fzj.unicore.uas.ft.http.AccessURLDocument;
import de.fzj.unicore.uas.fts.FileTransferImpl;
import de.fzj.unicore.uas.fts.byteio.ByteIOSizeResourceProperty;
import de.fzj.unicore.uas.fts.byteio.RandomByteIO;
import de.fzj.unicore.uas.impl.sms.SMSBaseImpl;
import de.fzj.unicore.uas.util.LogUtil;
import de.fzj.unicore.wsrflite.Kernel;
import de.fzj.unicore.wsrflite.security.ISecurityProperties;
import de.fzj.unicore.wsrflite.utils.Utilities;
import de.fzj.unicore.wsrflite.xmlbeans.WSUtilities;
import de.fzj.unicore.wsrflite.xmlbeans.rp.ImmutableResourceProperty;

/**
 * "Baseline" file transfer, which exposes a file using HTTP(s) via Jetty<br/>
 * 
 * The URI for download is "cryptographically hard to guess" and
 * is exposed using a WSRF resource property, thus protected
 * by the usual U6 access control.<br/>
 * 
 * Based on an idea by Roger Menday. <br/>
 * 
 * @author schuller
 * @since 1.0.1
 */
public class HttpFileTransferImpl extends FileTransferImpl{
	
	private static final Logger logger=LogUtil.getLogger(LogUtil.SERVICES,HttpFileTransferImpl.class);
	
	public final static QName RPAccessURL=AccessURLDocument.type.getDocumentElementName();
	
	@Override
	public void initialise(String name, Map<String, Object> map)throws Exception{
		super.initialise(name, map);
		try{
			if(source!=null){
				//TODO should have generic size property --- no need for RByteio here
				properties.put(RandomByteIO.RPSize ,new ByteIOSizeResourceProperty(this));
			}
			FileServlet.initialise();
			String url=exposeFile();
			AccessURLDocument urlRP=AccessURLDocument.Factory.newInstance();
			urlRP.setAccessURL(url);
			properties.put(RPAccessURL, new ImmutableResourceProperty(urlRP));
			setOK();
		}catch(Exception e){
			LogUtil.logException("Error initialising filetransfer",e,logger);
			status=STATUS_FAILED;
			description="Error initialising filetransfer";
			throw e;
		}
	}
	
	
	
	/**
	 * this adds the http(s) specific RPs to the Unicore Filetransfer RP doc <br/>
	 */
	@Override
	public XmlObject getResourcePropertyResponseDocument() throws Exception {
	
		XmlObject ftRPdoc=super.getResourcePropertyResponseDocument();
		WSUtilities.append(properties.get(RPAccessURL).getXml(),ftRPdoc);

		//size
		if(isExport()){
			SizeDocument sd=SizeDocument.Factory.newInstance();
			sd.setSize(BigInteger.valueOf(getDataSize()));
			WSUtilities.append(sd,ftRPdoc);
		}
		
		return ftRPdoc;
	}
	
	private String exposeFile()throws IOException{
		unique=Utilities.newUniqueID();
		String path=(isExport? source : target);
		FileServlet.exposeFile(unique, path, !overWrite,this.getStorageAdapter());
		accessURL=makeAccessURL(unique);
		return accessURL;
	}
	
	protected String makeAccessURL(String uniqueID){
		String add=null;
		if(!Boolean.parseBoolean(UAS.getProperty(SMSBaseImpl.SMS_DIRECT_FILETRANSFER))){
			//go via Gateway
			String base=Kernel.getKernel().getProperty(Kernel.WSRF_BASEURL);
			add=base.replace("services", "files")+"/"+uniqueID;
			if(logger.isDebugEnabled())logger.debug("Enabling HTTP access on URL: "+add);
		
		}
		else{
			//use UNICORE/X address
			String host=Kernel.getKernel().getProperty(Kernel.WSRF_HOST);
			String port=Kernel.getKernel().getProperty(Kernel.WSRF_PORT);
			String serv="/files";
			String proto="http";
			if("true".equalsIgnoreCase(Kernel.getKernel().getProperty(ISecurityProperties.WSRF_SSL))){
				proto="https";
			};
			add=proto+"://"+host+":"+port+serv+"/"+uniqueID;
		}
		if(Boolean.parseBoolean(UAS.getProperty("uas.filetransfers.http.preferPost"))){
			add=add+"?method=POST";
		}
		return add;
	}
	
	private void unExposeFile(){	
		FileServlet.unExposeFile(unique);
	}
	
	@Override
	public void destroy() {
		unExposeFile();
		super.destroy();
	}

	
	@Override
	protected Long getTransferredBytes() {
		Long t=FileServlet.getTransferredBytes(unique);
		if(t!=null){
			transferredBytes=t;
		}
		else{
			logger.debug("Can't get transferred bytes for transfer <"+unique+">");
		}
		return transferredBytes;
	}

}
