/*********************************************************************************
 * Copyright (c) 2008 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ********************************************************************************/
 
package de.fzj.unicore.uas.impl.tss.rp;

import org.apache.xmlbeans.XmlObject;
import org.unigrids.x2006.x04.services.tss.XgroupDocument;

import de.fzj.unicore.uas.impl.UASWSResourceImpl;
import de.fzj.unicore.uas.xnjs.XNJSFacade;
import de.fzj.unicore.wsrflite.impl.ResourceImpl;
import de.fzj.unicore.wsrflite.xmlbeans.ResourceProperty;
import eu.unicore.security.Client;
import eu.unicore.security.util.Log;

/**
 * Support publishing user's Unix groups as TSS/TSF properties. 
 * <p>
 * This RP publishes all 
 * groups which are permitted for the user (not the actually chosen groups). 
 * This depends on two factors: data from attribute sources and Perl TSI configuration option to enforce
 * OS groups. If present the TSI setting overrides the AS settings.
 * <p>
 * If TSI is not enforcing groups synchronization with OS then the list of available groups 
 * can be formed as a union of all groups allowed explicitly in AS and all OS groups for the uid.
 * If TSI is enforcing the synchronization then only the OS groups for the effective uid are 
 * allowed.
 * <p>
 * Note that result of the above algorithm depends on the actual uid which was selected and (if user
 * is allowed to use several uids) the result for the same grid user may vary. 
 */
public class XGroupsResourceProperty extends ResourceProperty<Object> {

	private static final long serialVersionUID = 1L;

	private XgroupDocument[] xdoc;
	
	public XGroupsResourceProperty(ResourceImpl parent) {
		super(parent);
	}

	@Override
	public XmlObject[] getXml() {
		return xdoc;
	}

	public XGroupsResourceProperty update(){
		getGroups();
		return this;
	}

	private String[] getAllowableGroups() {
		Client client=((UASWSResourceImpl)parentWSResource).getClient();
		if (client == null || client.getXlogin()==null || client.getXlogin().getUserName()==null)
			return new String[0];
		
		String xnjsReference=((UASWSResourceImpl)parentWSResource).getXNJSReference();
		try{
			return XNJSFacade.get(xnjsReference).getTSI(client).getGroups();
		}catch(Exception ex){
			Log.logException("Can't get groups from the operating system",ex);
			return new String[0];
		}
	}
	
	private void getGroups(){
		String[] groups = getAllowableGroups();
		xdoc=new XgroupDocument[groups.length];
		int i=0;
		for(String xg: groups){
			XgroupDocument xGroupDoc=XgroupDocument.Factory.newInstance();
			xGroupDoc.setXgroup(xg);
			xdoc[i]=xGroupDoc;
			i++;
		}
	}
}
