package de.fzj.unicore.uas.xnjs;

import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.unigrids.services.atomic.types.GridFileType;
import org.unigrids.services.atomic.types.PermissionsType;
import org.unigrids.services.atomic.types.ProtocolType;
import org.unigrids.x2006.x04.services.sms.ExportFileDocument;
import org.unigrids.x2006.x04.services.sms.ExportFileResponseDocument;

import de.fzj.unicore.uas.UAS;
import de.fzj.unicore.uas.client.FileTransferClient;
import de.fzj.unicore.uas.client.StorageClient;
import de.fzj.unicore.uas.client.FileTransferClient.IMonitorable;
import de.fzj.unicore.uas.fts.ProgressListener;
import de.fzj.unicore.uas.util.LogUtil;
import de.fzj.unicore.uas.util.Pair;
import de.fzj.unicore.wsrflite.Kernel;
import de.fzj.unicore.xnjs.Configuration;
import de.fzj.unicore.xnjs.ems.ExecutionException;
import de.fzj.unicore.xnjs.io.IStorageAdapter;
import de.fzj.unicore.xnjs.io.Permissions;
import de.fzj.unicore.xnjs.io.XnjsFile;
import de.fzj.unicore.xnjs.io.XnjsFileWithACL;

/**
 * Base class for UNICORE file imports<br/>
 * Sub classes need only provide an appropriate client class and 
 * the protocol that is used.
 *  
 * @author demuth
 * @author schuller
 */
public abstract class U6FileImportBase extends U6FileTransferBase {

	private GridFileType remote;

	public U6FileImportBase(Configuration configuration){
		super(configuration);
	}
	
	public final void run(){
		try{
			initSecurityProperties();

			StorageClient sms=new StorageClient(smsEPR,sec);

			localFile=target;

			//if overwrite policy is DONT_OVERWRITE: check if file exists
			if(OverwritePolicy.DONT_OVERWRITE.equals(overwrite)){
				XnjsFileWithACL f=null;
				try{
					f=getStorageAdapter().getProperties(localFile);
				}catch(ExecutionException ex){/*ignored*/}
				if(f!=null){
					throw new IOException("Target file <"+localFile+"> exists, and we have been asked to not overwrite.");
				}
			}
			
			long start=System.currentTimeMillis();

			remote = sms.listProperties(source);

			status=Status.RUNNING;
			if(remote.getIsDirectory())
			{
				transferFolder(remote, sms, target);
			}
			else
			{
				dataSize=remote.getSize();
				createParentDirectories(target);
				transferFile(remote, sms, target);
				copyPermissions();
			}
			status=Status.DONE;

			long end=System.currentTimeMillis();
			long l=end-start;
			computeMetrics(l);
			destroyFileTransferResource();
		}
		catch(ProgressListener.CancelledException ce){
			String msg="File transfer cancelled.";
			status=Status.DONE; //TODO: CANCELLED does not exist
			statusMessage=msg;
		}
		catch(Exception ex){
			String msg="Error executing filetransfer: "+ex.getMessage();
			LogUtil.logException(msg,ex,logger);
			status=Status.FAILED;
			statusMessage=msg;
		}

		onFinishCleanup();
	}

	@Override
	protected void onFinishCleanup(){
		super.onFinishCleanup();
		remote=null;
	}
	
	protected void computeMetrics(long consumedMillis)
	{
		long dataSize = getDataSize();
		float r=(float)dataSize/(float)consumedMillis;
		if(logger.isDebugEnabled()){
			logger.debug("Imported "+dataSize+" bytes in "+consumedMillis+" milliseconds, data rate= "+r + "kB/s");
		}
	}

	/**
	 * import the file. Will check if the remote file system is in fact the local one, and
	 * will invoke the transferFileByCopy() in that case
	 * @param source
	 * @param sms
	 * @param target
	 * @throws Exception
	 */
	protected void transferFile(GridFileType source, StorageClient sms, String target) throws Exception
	{
		//handle directory as target
		localFile = target;
		IStorageAdapter adapter=getStorageAdapter();
		XnjsFile xFile=adapter.getProperties(localFile);
		
		if(xFile!=null){
			if(xFile.isDirectory()){
				String name=getFileName(source.getPath());
				if(!target.endsWith(adapter.getFileSeparator())){
					localFile=target+adapter.getFileSeparator()+name;	
				}
				else localFile=target+name;
			}
			if(!xFile.getPermissions().isWritable()){
				throw new IOException("Local target file is not writeable");
			}
		}
		
		
		String remoteFSID=sms.getFileSystem().getDescription();
		String localFSID=getStorageAdapter().getFileSystemIdentifier();
		if(remoteFSID!=null && localFSID!=null && remoteFSID.equals(localFSID) && 
				!Kernel.getKernel().getBooleanProperty(UAS.SMS_TRANSFER_FORCEREMOTE, false)){
			transferFileLocal(source, sms, target);
		}
		else{
			transferFileFromRemote(source, sms);
		}
	}


	protected void transferFileLocal(GridFileType source, StorageClient sms, String target)throws Exception{
		String base=sms.getFileSystem().getMountPoint();
		//in some cases we have to fallback to the remote transfer
		if(base==null || (!isOverwrite() && source.getSize()>0)){
			transferFileFromRemote(source, sms);
		}
		String src=base+source.getPath();
		logger.info("Optimization enabled: importing file by local copy of <"+src+"> to <"+localFile+">");
		IStorageAdapter s=getStorageAdapter();
		String sRoot=s.getStorageRoot();
		try{
			s.setStorageRoot("/");
			s.cp(src, workdir+"/"+localFile);
			transferredBytes+=source.getSize();
		}
		finally{
			s.setStorageRoot(sRoot);
		}
		
	}

	
	protected void transferFileFromRemote(GridFileType source, StorageClient sms)throws Exception{
		ExportFileDocument req=ExportFileDocument.Factory.newInstance();
		req.addNewExportFile().setSource(source.getPath());
		req.getExportFile().setProtocol(ProtocolType.Enum.forString(getProtocol()));
		Map<String,String>ep=getExtraParameters();
		if(ep!=null && ep.size()>0){
			req.getExportFile().setExtraParameters(convert(ep));
		}
		ExportFileResponseDocument res=sms.ExportFile(req);
		fileTransferInstanceEpr=res.getExportFileResponse().getExportEPR();
		ftc=getFTClient();
		if(ftc instanceof FileTransferClient.IMonitorable){
			((IMonitorable)ftc).setProgressListener(this);
		}
		doRun();
	}
	
	protected void transferFolder(GridFileType source, StorageClient sms, String target) throws Exception
	{
		List<Pair<GridFileType,String>> collection = new ArrayList<Pair<GridFileType,String>>();
		dataSize = collectRemoteFiles(collection, source, sms, target);

		for (Pair<GridFileType,String> pair : collection) {

			GridFileType currentSource = pair.getM1();
			String currentTarget = pair.getM2();
			if(currentSource.getIsDirectory())
			{
				createParentDirectories(currentTarget+"/file");
				transferredBytes++;
			}
			else transferFile(currentSource, sms, currentTarget);
		}
	}



	/**
	 * Recursively gather all files and directories that need to be copied
	 */
	protected long collectRemoteFiles(List<Pair<GridFileType,String>> collection, GridFileType sourceFolder, StorageClient sms, String targetFolder) throws Exception
	{
		long result = 1;
		collection.add(new Pair<GridFileType,String>(sourceFolder, targetFolder));
		for (GridFileType child : sms.listDirectory(sourceFolder.getPath())) {
			String relative = child.getPath().substring(sourceFolder.getPath().length()+1);
			String target = targetFolder+getStorageAdapter().getFileSeparator()+relative;

			if(child.getIsDirectory()) 
			{
				result += collectRemoteFiles(collection, child, sms, target);
			}
			else 
			{
				collection.add(new Pair<GridFileType,String>(child,target));
				result+= child.getSize();
			}
		}
		return result;
	}

	public long getDataSize()
	{
		return dataSize;
	}

	protected void doRun()throws Exception{
		OutputStream os=null;
		boolean append=OverwritePolicy.APPEND.equals(overwrite);
		os=getStorageAdapter().getOutputStream(localFile, append);
		try{
			ftc.readAllData(os);
		}finally{
			if(os!=null)os.close();
		}
	}	
	
	protected void copyPermissions()throws Exception{
		PermissionsType p=remote.getPermissions();
		Permissions perm=XNJSFacade.getXNJSPermissions(p);
		getStorageAdapter().chmod(localFile, perm);
		if(logger.isDebugEnabled()){
			logger.debug("Copied remote permissions to :"+perm);
		}
	}


}
