package de.fzj.unicore.uas.testsuite;
import static org.junit.Assert.assertTrue;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Calendar;
import java.util.Random;

import junit.framework.Assert;

import org.ggf.schemas.jsdl.x2005.x11.jsdl.ApplicationType;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.JobDefinitionDocument;
import org.unigrids.services.atomic.types.ProtocolType;
import org.unigrids.x2006.x04.services.sms.ExportFileDocument;
import org.unigrids.x2006.x04.services.sms.ExportFileResponseDocument;
import org.unigrids.x2006.x04.services.sms.ImportFileDocument;
import org.unigrids.x2006.x04.services.sms.ImportFileResponseDocument;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.uas.UAS;
import de.fzj.unicore.uas.client.HttpFileTransferClient;
import de.fzj.unicore.uas.client.JobClient;
import de.fzj.unicore.uas.client.RByteIOClient;
import de.fzj.unicore.uas.client.StorageClient;
import de.fzj.unicore.uas.client.TSSClient;
import de.fzj.unicore.wsrflite.Kernel;
import de.fzj.unicore.wsrflite.utils.StopWatch;
import eu.unicore.bugsreporter.annotation.FunctionalTest;

/**
 * runs http filetransfer tests on a newly created USspace<br/>
 * 
 * @author schuller
 */
public class HttpTest extends AbstractJobRun{

	String url;
	EndpointReferenceType tssepr;

	StorageClient sms;

	@Override
	@FunctionalTest(id="HTTPTest", description="Tests the HTTP file transfer")
	public void testRunJob()throws Exception{
		url=Kernel.getKernel().getProperty(Kernel.WSRF_BASEURL);
		tssepr=EndpointReferenceType.Factory.newInstance();
		tssepr.addNewAddress().setStringValue(url+"/"+UAS.TSS+"?res=default_targetsystem");
		TSSClient tss=new TSSClient(tssepr,Kernel.getKernel().getSecurityProperties());
		Calendar c=tss.getCurrentTime();
		assertTrue(c!=null);
		runJob(tss);
	}


	@Override
	protected JobDefinitionDocument getJob() {
		JobDefinitionDocument jdd=JobDefinitionDocument.Factory.newInstance();
		ApplicationType app=jdd.addNewJobDefinition().addNewJobDescription().addNewApplication();
		app.setApplicationName("Date");
		app.setApplicationVersion("1.0");
		return jdd;
	}


	@Override
	protected void onFinish(JobClient jms) throws Exception {
		sms=job.getUspaceClient();

		// data stage in with perf measurement
		RByteIOClient importClient=sms.getImport("bigfile");
		int size=128000;
		int loops=10;
		byte[] data=new byte[size];
		Random r=new Random();

		StopWatch sw=new StopWatch();
		sw.start("Start import of "+size*loops/(1024*1024)+ " MBytes");
		System.out.println("Start import of "+size*loops/(1024*1024)+ " MBytes");
		for (int i = 0; i < loops; i++) {
			r.nextBytes(data);
			importClient.append(data);
		}
		sw.stop("End.");
		long time=sw.getCurrentTotal();
		System.out.println("End of data import, time="+time/1000+ " sec");
		System.out.println("Rate="+(size*loops)/time+ " kB/sec.");

		// data export
		RByteIOClient exportClient=sms.getExport("bigfile");

		sw.start("Start export of "+size*loops/(1024*1024)+ " MBytes");
		System.out.println("Start export of "+size*loops/(1024*1024)+ " MBytes");
		exportClient.readAllData(new OutputStream(){
			public void write(int b) throws IOException {
			}
		});

		sw.stop("End.");
		time=sw.getCurrentTotal();
		System.out.println("End of rbyteio data export, time="+time/1000+ " sec");
		System.out.println("Rate="+(size*loops)/time+ " kB/sec.");


		sw.start("Start Baseline/HTTP.");
		ExportFileDocument in=ExportFileDocument.Factory.newInstance();
		in.addNewExportFile().setProtocol(ProtocolType.BFT);
		in.getExportFile().setSource("/bigfile");
		ExportFileResponseDocument res=sms.ExportFile(in);

		HttpFileTransferClient ft=new HttpFileTransferClient(res.getExportFileResponse().getExportEPR(),
				Kernel.getKernel().getSecurityProperties());

		ft.readAllData(new OutputStream(){
			public void write(int b) throws IOException {
			}
		});


		sw.stop("End HTTP.");
		time=sw.getCurrentTotal();
		System.out.println("End of http data export, time="+time/1000+ " sec");
		System.out.println("Rate="+(size*loops)/time+ " kB/sec.");

		//test putting a file

		String testString="this is a test";
		ByteArrayInputStream bis=new ByteArrayInputStream(testString.getBytes());
		ImportFileDocument imp=ImportFileDocument.Factory.newInstance();
		imp.addNewImportFile().setProtocol(ProtocolType.BFT);
		imp.getImportFile().setDestination("/testfile");

		ImportFileResponseDocument res2=sms.ImportFile(imp);
		HttpFileTransferClient http2=new HttpFileTransferClient(
				res2.getImportFileResponse().getImportEPR(),
				Kernel.getKernel().getSecurityProperties());
		http2.writeAllData(bis);

		//and read it again
		ByteArrayOutputStream os=new ByteArrayOutputStream();
		sms.getExport("/testfile").readAllData(os);
		System.out.println(os.toString());
		assertTrue(os.toString().equals(testString));
		
		
		//test error: export non existent file
		try{
			sms.getExport("no such file", ProtocolType.BFT);
			Assert.fail("Expected exception since file does not exist");
		}catch(Exception e){
			assertTrue(e.getMessage().contains("FileNotFound"));
		}
		
	}


}
