/*********************************************************************************
 * Copyright (c) 2008 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *********************************************************************************/
 

package eu.unicore.genii.gffs;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.URI;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;

import de.fzj.unicore.uas.UAS;
import de.fzj.unicore.xnjs.Configuration;
import de.fzj.unicore.xnjs.ems.Action;
import de.fzj.unicore.xnjs.io.IFileTransfer;
import de.fzj.unicore.xnjs.io.XnjsFile;
import de.fzj.unicore.xnjs.io.simple.FTPTransferBase;
import de.fzj.unicore.xnjs.tsi.TSI;
import de.fzj.unicore.xnjs.util.AsyncCommandHelper;
import de.fzj.unicore.xnjs.util.LogUtil;
import de.fzj.unicore.xnjs.util.ResultHolder;
import eu.unicore.genii.gffs.GFFSTransfer.Lock;
import eu.unicore.security.Client;

/**
 * gsiftp file import implementation
 * 
 * TODO file size & transferred bytes
 * 
 * @author schuller
 */
public class GFFSUpload extends GFFSTransfer implements IFileTransfer {
	
	
	// do not try to create the genII state dir multiple times for multiple
	// stage-outs of the same job 
	private static Map<String,Lock> locks = new ConcurrentHashMap<String, Lock>();
	
	private Status status;
	private String statusMessage;
	private final URI target;
	private final String source;
	private final Client client;
	private final String uuid;
	private String cmd;
	private final Configuration configuration;
	AsyncCommandHelper ach;
	
	public GFFSUpload(Client client, String source,URI target, Configuration config) {
		this.configuration=config;
		status=Status.CREATED;
		this.target=target;
		this.source=source;
		this.client=client;
		uuid=UUID.randomUUID().toString();
		
	}
	
	public String getUniqueId(){return uuid;}
	
	public Status getStatus() {
		return status;
	}

	public String getStatusMessage() {
		if(statusMessage!=null)return statusMessage;
		return status!=null?status.toString():"<unknown>";
	}
	
	public long getTransferredBytes() {
		return transferredBytes;
	}

	public long getDataSize() {
		return -1;
	}

	/**
	 * does the ftp download
	 */
	public void run() {
		try{
			
			Action action=configuration.getInternalManager().getAction(getParentActionID());
			TSI tsi = configuration.getTargetSystemInterface(client);
			tsiFilesep = tsi.getFileSeparator();
			String jobId = action.getUUID();

			Lock lock;
			synchronized (locks) {
				lock = locks.get(jobId);
				if(lock == null)
				{
					lock = new Lock();
					locks.put(jobId,lock);
				}
			}
			
			synchronized (lock) {
				if(!lock.stateDirCreated)
				{
					createGenIIStateDir(client,configuration);
					lock.stateDirCreated = true;
				}
			}

			String parent = action.getExecutionContext().getWorkingDirectory();
			String stateDir = parent+tsiFilesep+GENII_STATE_DIR;
			
			String installationPath = UAS.getProperty(PROP_GENII_INSTALLATION_PATH_TSI);
			if(installationPath == null)
			{
				installationPath = configuration.getIDB().getProperty(PROP_GENII_INSTALLATION_PATH_TSI,null);
			}
			if(installationPath == null) installationPath = "";
			
			String executable = UAS.getProperty(PROP_GENII_EXECUTABLE);
			if(executable == null)
			{
				executable = configuration.getIDB().getProperty(PROP_GENII_EXECUTABLE,null);
			}
			if(executable == null) executable = GENII_EXECUTABLE;
			
			String params = "";
			

			if(OverwritePolicy.OVERWRITE.equals(getOverwritePolicy()))
			{
				cmd = installationPath + tsiFilesep + executable+" cp "+params+" 'local:"+source+"' '"+target+"'";
		
			}
			else if(OverwritePolicy.DONT_OVERWRITE.equals(getOverwritePolicy()))
			{
//				params += " -n";
				cmd = installationPath + tsiFilesep + executable+" cp "+params+" 'local:"+source+"' '"+target+"'";
			}
			
			
			ach=new AsyncCommandHelper(configuration,cmd,uuid,getParentActionID(),client);
			ach.setEnvironmentVariable(GENII_STATE_DIR_VARIABLE, stateDir);
			ach.submit();
			while(!ach.isDone()){
				Thread.sleep(2000);
			}
			ResultHolder res=ach.getResult();
			if(res.getExitCode()==null || res.getExitCode()!=0){
				statusMessage="GFFS upload failed.";
				try{
					String error=res.getStdErr();
					if(error!=null)statusMessage+=" Error details: "+error;
				}catch(IOException ex){
					LogUtil.logException("Could not read GFFS stderr",ex,logger);
				}
				status=Status.FAILED;
			}
			else{
				status=Status.DONE;
			}
		}catch(Exception ex){
			status=Status.FAILED;
			logger.error("Could not do GSSF upload",ex);
		}
	}

	public String getSource() {
		return source;
	}

	public String getTarget() {
		return target.toString();
	}

	
	
	public void update(XnjsFile info){
		if(info!=null){
			transferredBytes=info.getSize();
		}
	}
	
	
	public String getCommand(){
		return cmd;
	}
	
	public ResultHolder getResult(){
		return ach.getResult();
	}
	
	public String getProtocol(){
		return "rns";
	}
}
