package eu.unicore.genii.security;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.Serializable;
import java.security.cert.CertPath;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.codehaus.xfire.MessageContext;
import org.codehaus.xfire.handler.AbstractHandler;
import org.codehaus.xfire.handler.Phase;
import org.jdom.Element;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;
import org.xml.sax.InputSource;

import de.fzj.unicore.uas.util.LogUtil;
import de.fzj.unicore.wsrflite.security.SecurityManager;
import de.fzj.unicore.wsrflite.security.util.AttributeHandlingCallback;
import edu.virginia.vcgr.genii.client.context.CallingContextImpl;
import edu.virginia.vcgr.genii.client.context.CallingContextUtilities;
import edu.virginia.vcgr.genii.client.context.ICallingContext;
import edu.virginia.vcgr.genii.client.security.authz.AuthZSecurityException;
import edu.virginia.vcgr.genii.client.security.credentials.GIICredential;
import edu.virginia.vcgr.genii.client.security.credentials.assertions.DelegatedAssertion;
import edu.virginia.vcgr.genii.client.security.credentials.assertions.SignedAssertion;
import edu.virginia.vcgr.genii.client.ser.ObjectDeserializer;
import edu.virginia.vcgr.genii.context.ContextType;
import eu.unicore.genii.GenesisConstants;
import eu.unicore.security.SecurityTokens;
import eu.unicore.security.xfireutil.AuthInHandler;

public class GAMLAssertionInHandler extends AbstractHandler implements GenesisConstants {

	private static final Logger logger = LogUtil.getLogger(LogUtil.SECURITY,GAMLAssertionInHandler.class);

	public GAMLAssertionInHandler(){
		setPhase(Phase.POLICY);
		after(AuthInHandler.class.getName());
		addAttributeCallback();
	}

	//add callback that extracts attributes when creating a client
	private void addAttributeCallback(){
		SecurityManager.addCallback(getCallback());
	}


	/**
	 * This method retrieves GenesisII's GAML Assertion from the SOAP Header, generates an unprivileged certificate, 
	 * and puts this information into the the security context for running jobs with GenesisII based data stagings.  
	 * 
	 */
	public void invoke(MessageContext ctx) throws Exception
	{	

		SecurityTokens securityTokens = (SecurityTokens) ctx.getProperty(
				SecurityTokens.KEY);
		if (securityTokens == null)
		{
			throw new Exception("No security info in headers. Wrong configuration: " +
					AuthInHandler.class.getCanonicalName() + " handler" +
			" must be configure before this handler.");

		}


		// get the SOAP header
		Element header = ctx.getInMessage().getHeader();
		if (header == null){
			return ;	
		} 

		// get gaml header extensions
		@SuppressWarnings("rawtypes")
		List extensions = header.getChildren(GENII_GAMLASSERTION, GENII_NS);
		if(extensions.size()==0){
			logger.debug("No GAML Assertion header found.");
			return;
		}
		Element gamlElement = (Element) extensions.get(0);		
		ByteArrayOutputStream bos=new ByteArrayOutputStream();
		new XMLOutputter(Format.getPrettyFormat()).output(gamlElement, bos);

		String gamlAsString = bos.toString(); 
		ByteArrayInputStream in = new ByteArrayInputStream(gamlAsString.getBytes());
		InputSource source = new InputSource(in);

		ContextType ct = (ContextType) ObjectDeserializer.deserialize(
				source, ContextType.class);
		ICallingContext callingContext =
			CallingContextUtilities.setupCallingContextAfterCombinedExtraction(
					new CallingContextImpl(ct));
		Collection<Serializable> signedAssertions = (Collection<Serializable>)
			callingContext.getTransientProperty(GIICredential.CALLER_CREDENTIALS_PROPERTY);
		validateAssertions(securityTokens,signedAssertions);
		CertPath user = determineUser(signedAssertions);
		if(user == null)
		{
			//				throw new Exception("The correct user certificate could not be determined.");
		}
		else
		{
			securityTokens.setUser(null);
		}



		if(logger.isDebugEnabled()){
			logger.debug("Extracted GAML header:\n"+gamlAsString);		
		}

		// get the security context
		SecurityTokens tokens=((SecurityTokens)ctx.getProperty(SecurityTokens.KEY));		

		if (tokens==null){
			//should never happen in production (always have AuthInHandler)
			logger.warn("No security tokens found, new ones created...");			
			tokens=new SecurityTokens();
			ctx.setProperty(SecurityTokens.KEY, tokens);
		}

		tokens.getContext().put(GENII_GAMLASSERTION_REF, callingContext);
	}

	private CertPath determineUser(Collection<Serializable> signedAssertions)
	{

		if (signedAssertions != null) 
		{
			Iterator<Serializable> itr = signedAssertions.iterator();
			while (itr.hasNext()) 
			{
				GIICredential cred = (GIICredential) itr.next();
				if (cred instanceof SignedAssertion)
				{

					SignedAssertion signedAssertion = (SignedAssertion) cred;
					while(signedAssertion != null)
					{
						X509Certificate[] identity = signedAssertion.getAuthorizedIdentity();
						if(identity == null)
						{
							try {
								CertificateFactory cf = CertificateFactory.getInstance("X.509");
								return cf.generateCertPath(Arrays.asList(identity));
							} catch (CertificateException e) {
								logger.warn("Unable to build CertPath from Certificate Array "+e.getMessage(), e);
								return null;
							}
						}
						if(signedAssertion instanceof DelegatedAssertion)
						{
							DelegatedAssertion delegated = (DelegatedAssertion) signedAssertion;
							signedAssertion = delegated.unwrap();
						}
						else
						{
							signedAssertion = null;
						}
					}
				}
			}
		}
		return null;
	}

	private void validateAssertions(SecurityTokens securityTokens, Collection<Serializable> signedAssertions) throws Exception
	{

		Certificate serverCert = SecurityManager.getServerCert();
		// Deserialize the encoded caller-credentials
		if (signedAssertions != null) 
		{

			Certificate consignorCert = securityTokens.getConsignorCertificate();

			Iterator<Serializable> itr = signedAssertions.iterator();
			while (itr.hasNext()) 
			{
				GIICredential cred = (GIICredential) itr.next();
				if (cred instanceof SignedAssertion)
				{
					// Holder-of-key token
					SignedAssertion signedAssertion = (SignedAssertion) cred;

					// Check validity and verify integrity
					signedAssertion.checkValidity(new Date());
					signedAssertion.validateAssertion();


					// If the assertion is pre-authorized for us, unwrap one
					// layer
					if ((serverCert != null)
							&& (signedAssertion.getAuthorizedIdentity()[0]
							                                            .equals(serverCert)))
					{
						if (!(signedAssertion instanceof DelegatedAssertion))
						{
							throw new AuthZSecurityException(
									"GAML credential \""
									+ signedAssertion
									+ "\" does not match the incoming message sender");
						}
						signedAssertion =
							((DelegatedAssertion) signedAssertion).unwrap();
					}

					// Verify that the request message signer is the same as the
					// one of the holder-of-key certificates

					if (!consignorCert.equals(signedAssertion.getAuthorizedIdentity()[0])) {
						throw new AuthZSecurityException(
								"GII credential \""
								+ signedAssertion
								+ "\" does not match the incoming message sender");
					}
				}
			}
		}
	}

	private static AttributeHandlingCallback aac;

	private synchronized AttributeHandlingCallback getCallback(){
		if(aac==null){
			aac=new GAMLAssertionAttributeCallback();
		}
		return aac;
	}




	/**
	 * callback functions
	 */
	private static class GAMLAssertionAttributeCallback implements AttributeHandlingCallback{
		/**
		 * gets the gaml assertion and an unprivileged cert from the sec tokens and returns it as an attribute
		 * which will be stored in the client
		 */
		public Map<String,Serializable> extractAttributes(SecurityTokens tokens) {
			Map<String,Serializable> result = new HashMap<String,Serializable>();

			String gaml= (String)tokens.getContext().get(GENII_GAMLASSERTION);

			if(gaml!=null){
				result.put(GENII_GAMLASSERTION_REF , gaml);
			}
			return result;
		}
	}


}
