/*********************************************************************************
 * Copyright (c) 2006 Forschungszentrum Juelich GmbH
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 *
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its
 * contributors may be used to endorse or promote products derived from this
 * software without specific prior written permission.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ********************************************************************************/
package eu.unicore.uas.metadata;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.log4j.Logger;
import org.apache.tika.exception.TikaException;
import org.apache.tika.metadata.Metadata;
import org.apache.tika.parser.ParseContext;
import org.apache.tika.parser.Parser;
import org.apache.tika.sax.BodyContentHandler;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import de.fzj.unicore.uas.metadata.ExtractionStatistics;
import de.fzj.unicore.uas.metadata.MetadataManager;
import de.fzj.unicore.uas.util.LogUtil;
import de.fzj.unicore.wsrflite.Kernel;
import de.fzj.unicore.xnjs.ems.ExecutionException;
import de.fzj.unicore.xnjs.io.IStorageAdapter;
import de.fzj.unicore.xnjs.io.XnjsFile;
import eu.unicore.uas.metadata.MetadataFile.MD_State;
import java.util.Collections;

/**
 * Crawlers through the file system and collect metadata for files
 *
 * @author w.noor
 * @author schuller
 * @author jrybicki
 */
public class MetadataCrawler implements Callable<ExtractionStatistics> {

    private static final Logger LOG = LogUtil.getLogger(LogUtil.SERVICES, MetadataCrawler.class);
    //how long the crawling waits when scheduled (in seconds)
    public static final long DEFAULTSCHEDULEDELAY = 1;
    private final String base; /* base is absolute/relative path where crawler start its job.*/

    private final ContentHandler handler = new BodyContentHandler(-1);
    private final ParseContext parseContext = new ParseContext();
    private final Parser parser;
    private final int depthLimit; /* depthLimit defines the limit for the crawler.*/

    private final MetadataManager metadataManager;
    private final IStorageAdapter storage;
    private int currentLevel = 0;
    private AtomicInteger documentsProcessed = new AtomicInteger(0);
    static final String PARSERCLASSNAME = "uas.metadata.parser.class";

    /**
     * Crawls through resources available in basepath (and subdirs to @code{dephLimit}) via storage
     *<p>
     * Found resources are indexed via metadataManager. Extraction of the metadata is done
     * via parser (e.g. Apache Tika) defined in the config file.
     * </p>
     * 
     * @param metadataManager reference to the metadata manager (for creating metadata)
     * @param storage reference to the storage that should be crawled
     * @param base directory to be crawled
     * @param depthLimit depth of the crawling process
     * 
     * @throws InstantiationException thrown when parser instantiation fails
     * @throws IllegalAccessException  thrown when parser instantiation fails
     */
    @SuppressWarnings("unchecked")
    public MetadataCrawler(MetadataManager metadataManager, IStorageAdapter storage, String base, int depthLimit) throws InstantiationException, IllegalAccessException {

        this.base = base;
        this.depthLimit = depthLimit;
        this.metadataManager = metadataManager;
        this.storage = storage;

        Class<? extends Parser> parserClass = eu.unicore.uas.metadata.utils.ExtensionParser.class;
        try {
            String mpN = Kernel.getKernel().getProperty(PARSERCLASSNAME, "eu.unicore.uas.metadata.utils.ExtensionParser");
            if (mpN != null) {
                parserClass = (Class<? extends Parser>) Class.forName(mpN);
                LOG.info("Using Metadata Parser class: " + mpN);
            }

        } catch (Exception ex) {
            LOG.fatal("Unable to localize metadata parser class: " + parserClass, ex);
        } finally {
            parser = parserClass.newInstance();
        }

    }

    /**
     * Starts the crawling process
     * <p>
     * Process:
     * 1) extract a list of files from the storage
     * 2) check status changes (file added, file removed, etc)
     * 3) update md index 
     * 
     * 
     * @return Extraction Statistics
     */
    @Override
    public ExtractionStatistics call() {
        String fullBase = base;
        LOG.info("STARTING crawler from " + fullBase+" crawling depth "+depthLimit);

        long start = System.currentTimeMillis();
        List<String> fileList = new ArrayList<String>();

        try {
            getFiles(fullBase, fileList, depthLimit);
            
            Map<String, MD_State> list = statusCheck(fileList);
            for (Map.Entry<String, MD_State> entry : list.entrySet()) {
                
                switch (entry.getValue()) {
                    case CHK_CONSISTENCE:
                        //there is already a md file, update md
                        //XXX: we migth also use tika here (complementary)
                        Map<String, String> metadata = Collections.emptyMap();
                        metadataManager.updateMetadata(entry.getKey(), metadata);
                        break;
                    case NEW:
                        //it is new and no user metadata was created-> try extract
                        Map<String, String> extracted = extractMetadata(entry.getKey());
                        metadataManager.createMetadata(entry.getKey(), extracted);
                        break;
                    case RESOURCE_DELETED:
                        metadataManager.removeMetadata(entry.getKey());
                        break;
                    default:
                        //ignore for now?
//                        throw new IllegalArgumentException("State: "+resourceState+" is unknown");
                }

                documentsProcessed.incrementAndGet();
            }
        } catch (Exception ex) {
            LOG.warn("Error while crawling the metadata", ex);
        }


        long time = System.currentTimeMillis() - start;
        LOG.info("EXITING crawler on " + fullBase + " time " + time + " ms.");
        ExtractionStatistics stats = new ExtractionStatistics();
        stats.setDocumentsProcessed(documentsProcessed.get());
        stats.setDurationMillis(time);

        return stats;
    }

    /**
     * Filters list of files to detect:
     * -removal of resource files without removal of metadata files (md sould be removed)
     * -creation of resource files without metadata files (resource should be indexed)
     * -updates in metadata file (index should be updated)
     * 
     * FIXME:
     * -it might improve the performance to remove resource from file list when
     * resource.md file was found (and other way round). In some cases 2 speed-up
     * but concurrent modification exception.
     * -possible solution would be to have a callback in the listFiles method 
     * where such cases could be caught... further refactoring.
     * 
     * @param files
     * @return map of resource names with respective status (updated, removed, new)
     */
    protected static Map<String, MetadataFile.MD_State> statusCheck(List<String> files) {

        Map<String, MetadataFile.MD_State> statuses = new HashMap<String, MD_State>();

        for (String file : files) {
            String resource = null;
            if (MetadataFile.isMetadataFileName(file)) {
                resource = MetadataFile.getResourceName(file);
                if (statuses.containsKey(resource)) {
                    //overwrite NEW with CHK
                    statuses.put( resource, MD_State.CHK_CONSISTENCE);
                } else {
                    statuses.put(resource, MD_State.RESOURCE_DELETED);
                }

            } else {
                resource = file;
                if (statuses.containsKey(resource)) {
                    //overwrite DELETED with CHK
                    statuses.put(resource, MD_State.CHK_CONSISTENCE);
                } else {
                    statuses.put(resource, MD_State.NEW);
                }
            }
        }
        return statuses;
    }

    
    private void getFiles(String directoryName, List<String> list, int level) throws ExecutionException {
        currentLevel++;
        //TODO: make it more flexible and define extension to be avoided
        if (isLevelExceeded() || directoryName.contains(".svn")) {
            return;
        }

        XnjsFile x = storage.getProperties(directoryName);
        if (x != null && x.isDirectory()) {
            XnjsFile[] gridFiles = storage.ls(directoryName);
            for (int j = 0; j < gridFiles.length; j++) {
                XnjsFile x2 = storage.getProperties(gridFiles[j].getPath());
                if (x2.isDirectory()) {
                    getFiles(gridFiles[j].getPath(), list, level);
                } else {
                    list.add(gridFiles[j].getPath());
                }
            }
        }

        currentLevel--;
    }

    private boolean isLevelExceeded() {
        return currentLevel > depthLimit;
    }

    private Map<String, String> extractMetadata(String file) throws ExecutionException, IOException, SAXException, TikaException {
        Map<String, String> ret = new HashMap<String, String>();
        Metadata meta = new Metadata();
        meta.add(Metadata.RESOURCE_NAME_KEY, file);
        InputStream is = storage.getInputStream(file);
        parser.parse(is, handler, meta, parseContext);
        is.close();
        for (String key : meta.names()) {
            ret.put(key, meta.get(key));
        }
        return ret;
    }
}
